package PureMessage::Policy::Virus::Daemon;

use strict;
use ActiveState::ProcessPool;
use ActiveState::Milter::SockSpec;
use PureMessage::Socket;
use PureMessage::Config '%config';
use PureMessage::Logger;
use PureMessage::Policy::Virus::Config;
use PureMessage::Policy::Virus::Protocol qw(send recv);

use Data::Dump 'dump';

our $VERSION = '0.02';

$| = 1;

# Ignore SIGPIPE -- we detect errors via EOFs and such.
$SIG{PIPE} = 'IGNORE';

our %my_config;
* my_config = \%PureMessage::Policy::Virus::Config::config;

sub new {
    my $cls = shift;
    my $o = bless {
	port	    => $my_config{daemon}{port},
	sockspec    => [ActiveState::Milter::SockSpec::parse($my_config{daemon}{port})],
	start_time  => time,
    }, ref($cls) || $cls;

    # Figure out what backends to run (and how many)
    my %backends;
    for (grep $my_config{backend}{$_}{enabled}, keys %{$my_config{backend}}) {
        my $b_end = $_;
	if ($my_config{backend}{$b_end}{daemon}) {
	    require PureMessage::_sys;
	    my $b = "PureMessage::Policy::Virus::$b_end";
	    eval { PureMessage::_sys::load_module($b) };
	    if ($@) {
		warn $@;
	    }
	    else {
		$backends{$b} = $b->daemon($my_config{backend}{$b_end});
	    }
	}
    }
    $o->{bk} = \%backends;
    $o->{backends} = [sort keys %backends];
    $o;
}

sub nothing_to_do {
    my $o = shift;
    return not @{$o->{backends}};
}

sub daemon {
    my $o = shift;
    require PureMessage::_sys;
    chdir '/'			or die "can't chdir to /: $!";
    defined (my $pid = fork)	or die "can't fork: $!";
    exit if $pid;
    PureMessage::_sys::setsid()	or die "can't start new session: $!";
    my $log = "$config{log_dir}/vscan_out";
    open(STDOUT, ">> $log")	or die "can't log to $log: $!";
    open(STDERR, ">&STDOUT")	or die "can't dup STDERR 2>&1: $!";
}

sub init_listener {
    my $o = shift;
    my $sock = $o->{port};
    my ($scheme, $port, $if) = @{$o->{sockspec}};
    if ($scheme eq 'inet') {
	$o->{socket} = PureMessage::Socket::INET->new(
	    $if ? (LocalAddr => $if) : (),
	    LocalPort	=> $port,
	    Type	=> SOCK_STREAM,
	    ReuseAddr	=> 1,
	    Listen	=> 1024,
	) or die "Can't bind to socket $sock: $!";
    }
    elsif ($scheme eq 'unix') {
	if (-S $port) {
	    # if the socket file already exists, try to figure
	    # out if it is stale
	    my $s = PureMessage::Socket::UNIX->new(
	        Peer    => $port,
		Type	=> SOCK_STREAM,
	    );
	    unlink($port) unless $s;
	}
	$o->{socket} = PureMessage::Socket::UNIX->new(
	    Type	=> SOCK_STREAM,
	    Local	=> $port,
	    Listen	=> 1024,
	) or die "Can't bind to socket $sock: $!";
    }
    else {
	die "Unsupported socket scheme $scheme";
    }
}
sub fini_listener {
    my $o = shift;
    my ($scheme, $port, $if) = @{$o->{sockspec}};
    unlink $port if $scheme eq 'unix';
    delete $o->{socket};
}

sub init_scanners {
    my $o = shift;
    # A process pool large enough for "concurrency_limit" of each backend.
    $o->{pool} = ActiveState::ProcessPool->new(scalar $o->slots);
    $o->scanner_start($_) for $o->slots;
    $o->refresh_fds;
}
sub fini_scanners {
    my $o = shift;
    $o->scanner_stop($_) for $o->slots;
    $o->{pool} = undef;
}

sub init {
    my $o = shift;
    $o->{chatting} = []; # set of active client sockets
    $o->{pending} = []; # queue of pending scans
    $o->{running}[$_] = undef for $o->slots; # active scans
}

sub slots {
    my $o = shift;
    my $top = $o->{_top_} ||=
	$my_config{daemon}{concurrency_limit} * @{$o->{backends}};
    my @slots = (0 .. $top - 1);
}

sub log {
    my ($self, $level, $msg) = @_;
    $self->{logger}->log($level, $msg);
}

sub run {
    my $o = shift;

    $o->{logger} = PureMessage::Logger->new(
	ident => "vscan",
	log_to => "vscan_log",
    );

    $o->init_listener;
    $o->daemon;
    $o->init_scanners;
    $o->init;

    return unless $o->slots;

    my $scan_poll_interval = $my_config{daemon}{scan_poll_interval};

    $o->log("INFO", "vscan starting up with " . $o->slots . " slots");
    while (1) {

	# If the listening socket is shutdown, the queue is empty, and no
	# slots are in use, we're done. This is the result of somebody sending
	# a 'stop' request.
	last if $o->check_processing == 0;

	# Set up the read and write vectors. Choose carefully. For each slot,
	# we check if we're servicing something. If we are, we add the
	# relevant fds for the set.
	my ($rfds, $wfds) = $o->choose_fdsets;

        # Wait for something to become readable or writable.
        my $nfound = select($rfds, $wfds, undef, $scan_poll_interval);
	my $timenow = time;

	$o->check_dead_slots($timenow);

	# First service new requests coming in on the master socket.
	$o->service_listener if fd_isset($rfds, $o->{socket});

        # Walk through the active clients and process those ready to
        # talk. This means reading and writing packets. If a command can
        # be dealt with immediately (like 'ping') an answer packet is
        # immediately pushed onto the active clients list. If the
        # command is 'scan', the packet is moved from the chatting list
        # into the 'scanner queue', so that the next time a scanner
        # becomes available it can begin processing.
        for my $client ($o->chatting) {
            $o->read_sock($client) if fd_isset($rfds, $client->{sock});
            $o->write_sock($client) if fd_isset($wfds, $client->{sock});
        }
	@{$o->{chatting}} = grep { $_ && ! $_->{del} } @{$o->{chatting}};

	# Check for timeouts in queued scan requests.
	for my $client ($o->queued) {
	    $o->check_for_timeout(undef, $client, $timenow);
	}

        # Deal with active scans.
        for my $slot ($o->slots) {
            my $ent = $o->slot($slot);

            $o->recover_scanner($slot, $ent) if fd_isset($rfds, $o->efd($slot));
            next unless $o->slot_inuse($slot);

	    $o->check_for_timeout($slot, $ent, $timenow);
	    next unless $o->slot_inuse($slot);

            $o->write_scanner($slot, $ent) if fd_isset($wfds, $o->wfd($slot));
            next unless $o->slot_inuse($slot);

            $o->read_scanner($slot, $ent) if fd_isset($rfds, $o->rfd($slot));
            next unless $o->slot_inuse($slot);
        }

	# If there are any idle slots, fill them from the scanner queue:
        for my $slot ($o->slots_idle) {
            last unless $o->queuesize;
            $o->dequeue($slot);
        }
    }
    $o->log("INFO", "vscan shutting down");
}

sub check_processing {
    my $o = shift;
    my $sock     = defined $o->{socket} && $o->{socket}->sockname ? 1 : 0;
    my $chatting = $o->chatting;
    my $pending  = $o->queuesize;
    my $running  = $o->slots_inuse;
    return $sock + $chatting + $pending + $running;
}

sub choose_fdsets {
    my $o = shift;
    my (@rfds, @wfds);
    push @rfds, $o->{socket} if $o->{socket} && $o->{socket}->sockname;
    for my $ent ($o->chatting) {
        push @rfds, $ent->{sock} if (
            not $ent->{eof} and ($ent->{read_sock} or $ent->{write_sock})
        );
        push @wfds, $ent->{sock} if $ent->{write_sock};
    }
    for my $slot ($o->slots) {
        next if $o->slot_dead($slot);
        push @rfds, $o->efd($slot); # always check for scanner errors
        next unless $o->slot_inuse($slot);
        my $ent = $o->slot($slot);
        push @rfds, $o->rfd($slot) if $ent->{read_scanner};
        push @wfds, $o->wfd($slot) if $ent->{write_scanner};
    }
    my $rfds = fd_set(@rfds);
    my $wfds = fd_set(@wfds);
    ($rfds, $wfds);
}

sub check_for_timeout {
    my ($o, $slot, $ent, $timenow) = @_;
    my $scan_timeout = $my_config{daemon}{scan_limit};
    my $actv_timeout = $my_config{daemon}{inactive_limit};
    my $inactive = $timenow - ($ent->{read_scanner_last} || $timenow);
    my $elapsed  = $timenow - ($ent->{start} || $timenow);
    my $error;
    my $prob;
    if ($actv_timeout and $inactive > $actv_timeout) {
	$error = "Scanner spent too long being inactive"
	    if defined $slot;
	$prob = 'TIMEOUT:inactive_limit';
    }
    if ($scan_timeout and $elapsed >= $scan_timeout) {
	$error = "Scanner took too long processing the message";
	$prob = 'TIMEOUT:scan_limit';
    }
    if ($error) {
	$ent->{resp} = [
	    op	    => 'e',
	    problem => $prob,
	    descrip => $error,
	    command => $ent->{cmd},
	    file    => $ent->{file},
	    created => $ent->{create},
	    started => $ent->{start},
	    elapsed => $elapsed,
	    report  => $ent->{rbuf},
	];
	$ent->{read_scanner} = 0;
	$ent->{write_scanner} = 0;
	$ent->{write_sock} = 1;
	$ent->{read_sock} = 1;
	$ent->{restart} = 1;
	if (defined $slot) {
	    $o->log("INFO", "Timeout in scanner $slot: $error");
	    $o->free_slot($slot);
	}
	else {
	    $o->log("INFO", "TIMEOUT in queued message: $error");
	    $o->free_queued($ent);
	}
    }
}

sub service_listener {
    my $o = shift;
    my $client = $o->{socket}->accept or return;
    $client->blocking(0);
    $o->log("DEBUG", "Enqueuing connection ${$client}");
    $o->chat_insert($o->alloc_ent($client));
}

sub recover_scanner {
    my ($o, $slot, $ent) = @_;

    # The error fd became readable. This means the scanner exited for some
    # reason.  We never get this condition when calling restart(), because we
    # close the fd before killing the child. This condition is always an
    # error. The choices are: (a) if the child has served at least 1 message,
    # just restart it. It's probably just a scanner bug or some kind of denial
    # of service attack against it; (b) otherwise abort, making lots of noise
    # so the log fills up and alerts the administrator.

    my $used = $o->{pool}->used($slot);
    if ($used == 0) {
        my $pid = $o->{pool}->pid($slot);
	my $name = $o->slot_name($slot);
	$o->log("INFO", "Scanner $slot ($name, $pid) died " .
			"without processing any messages");
    }

    # If the slot is in use, its use count is >= 1, so we'll just let it
    # be restarted when we've sent the error packet to PureMessage.
    if ($o->slot_inuse($slot)) {
        $o->err_scanner_closed($slot, $ent);
    }
    # If the slot was used at least once, but then died in between messages,
    # it's kinda weird, but we'll just restart it and go on.
    elsif ($used) {
        $o->scanner_restart($slot, 1);
        $o->refresh_fds;
    }
    # If the scanner died before being used even once, it will be stopped and
    # the slot marked as dead. We'll continue processing data with the
    # remaining children, but when all of them run out, we'll stop.
    else {
        $o->scanner_stop($slot, 1);
        $o->refresh_fds;
        $o->slot_dead($slot, 1);

	# Check whether the last scanner just died. If so, the accept loop
	# won't allow 'scan' requests to be queued; but those requests already
	# in the queue must be removed from the queue and explicitly failed.
	if ($o->slots_dead == $o->slots) {
	    while (my $ent = $o->popqueue) {
		$ent->{resp} = [
		    op	    => 'e',
		    problem => 'SCANNERS_DEAD',
		    descrip => 'All scanners died while the request was queued',
		    command => $ent->{cmd},
		    file    => $ent->{file},
		    created => $ent->{create},
		    started => 0,
		    elapsed => 0,
		    report  => '',
		];
		$ent->{read_scanner} = 0;
		$ent->{write_scanner} = 0;
		$ent->{write_sock} = 1;
		$ent->{read_sock} = 1;
		$ent->{restart} = 1;
	    }
	}
	$o->log("INFO", "Will restart slot $slot in " . 
			"$my_config{daemon}{respawn_wait}s");
    }
}

sub write_scanner {
    my ($o, $slot, $ent) = @_;
    $o->{pool}->send($slot, %{$ent->{send_scanner}});
    $ent->{write_scanner} = 0;
}

sub read_scanner {
    my ($o, $slot, $ent) = @_;
    my @kv = $o->{pool}->recv($slot);
    return $o->err_scanner_closed($slot, $ent) unless @kv;

    {
	my ($file, @v);
	my @cp = @kv;
	while (@cp) {
	    my ($k, $v) = splice(@cp, 0, 2);
	    if    ($k eq 'file') { $file = $v  }
	    elsif ($k eq 'v')    { push @v, $v }
	}
	if (@v) {
	    $o->log("DEBUG", "$file: infected with @v");
	}
	else {
	    $o->log("DEBUG", "$file: clean");
	}
    }

    $ent->{read_scanner_last} = time;
    $ent->{resp} = [
	@kv,
	op => $ent->{cmd},
	backend => $o->slot_name($slot),
    ];
    $ent->{read_scanner} = 0;
    $ent->{read_sock}   = 1;
    $ent->{write_sock}  = 1;
    $o->free_slot($slot);
}

#=============================================================================
# Handle chat from clients.
#=============================================================================
sub chatting {
    my $o = shift;
    @{$o->{chatting}};
}

sub chat_insert {
    my ($o, $client) = @_;
    push @{$o->{chatting}}, $client;
}

sub chat_free {
    my ($o, $client) = @_;
    my $fileno = $client->{sock}->fileno;
    $o->{chatting} = [(
	grep { $_ && $_->{sock}->fileno != $fileno }
	@{$o->{chatting}}
    )];
}

sub handle {
    my ($o, $cmd, $ent, %kv) = @_;
    $ent->{cmd} = $cmd;

    # Scan a file for viruses:
    if ($cmd eq 's' or $cmd eq 'c') {
	my $qs = $my_config{daemon}{queue_limit},
	my %ACT = (
	    s	    => 'scan',
	    c	    => 'clean',
	);
	my %err = (
	    problem => undef,
	    descrip => undef,
	    command => $cmd,
	    file    => $kv{file},
	    created => $ent->{create},
	    started => 0,
	    elapsed => 0,
	    report  => 'Request ignored',
	    op	    => 'e',
	);

	$o->log("DEBUG", "$ACT{$cmd} request for $kv{file}");

	if (!-e $kv{file}) {
	    $err{problem} = 'FILE';
	    $err{descrip} = 'Tried to scan nonexistent file',
	}
	elsif (!-f $kv{file}) {
	    $err{problem} = 'FILE';
	    $err{descrip} = 'Tried to scan non-file',
	}
	elsif (!-s $kv{file}) {
	    $err{problem} = 'FILE';
	    $err{descrip} = 'Tried to scan zero length file',
	}
	elsif ($qs and $o->queuesize >= $qs) {
	    $err{problem} = 'QUEUE_FULL';
	    $err{descrip} = "Scan queue is full ($qs entries)";
	}
	elsif ($o->slots_dead == $o->slots) {
	    $err{problem} = 'SCANNERS_DEAD',
	    $err{descrip} = 'All scanners are dead';
	}
	else {
	    $o->enqueue($ent);
	    $ent->{file} = $kv{file};
	    if($ACT{$cmd} eq 'clean'){		#CXmail change
		$ent->{send_scanner} = {
                    action      => $ACT{$cmd},
                    file        => $kv{file},
                };
	    }else{
	        $ent->{send_scanner} = {
		    action => $ACT{$cmd},
		    %kv,
	         };
	    }
	    $ent->{write_scanner} = 1;
	    $ent->{read_scanner} = 1;
	    $ent->{read_sock}   = 0; # wait until we're finished scanning
	}

	if ($err{problem}) {
	    $o->log("ERROR", "$err{descrip}");
            $ent->{resp} = [%err];
            $ent->{read_sock}  = 1;
            $ent->{write_sock} = 1;
        }
    }
    # Are we alive? Schedule a reply right away.
    elsif ($cmd eq 'p') {
      PING_RESPONSE:
        $o->log("DEBUG", "received PING request");
	my %hash = (
	    op		=> 'p',
	    pid		=> $$,
	    numScanners	=> $my_config{daemon}{concurrency_limit},
	    connections	=> scalar($o->chatting),
	    queued	=> scalar($o->queuesize),
	    slotsInUse	=> scalar($o->slots_inuse),
	    slotsIdle	=> scalar($o->slots_idle),
	    slotsDead	=> scalar($o->slots_dead),
	    start_time	=> $o->{start_time},
	);
	for my $slot ($o->slots) {
	    if ($o->slot_dead($slot)) {
		$hash{"slotUsed_$slot"} = 'DEAD';
		$hash{"slotPID_$slot"}  = 'DEAD';
	    }
	    else {
		$hash{"slotUsed_$slot"} = $o->{pool}->used($slot);
		$hash{"slotPID_$slot"}  = $o->{pool}->pid($slot);
	    }
	}
	$ent->{resp} = [%hash];
	$ent->{read_sock}  = 1;
	$ent->{write_sock} = 1;
    }
    # Stop. Shutdown the scanners and exit.
    # Gentle stop. Close the accept socket, but let everybody finish.
    elsif ($cmd eq 'x') {
	if ($kv{hard}) { # hard stop -- just exit
	    $o->log("INFO", "received *hard* stop request. Shutting down!");
	    $o->fini_scanners;
	    $o->fini_listener;
	    exit 0;
	}
	else {
	    $o->log("INFO", "received *soft* stop request. Closing socket!");
	    $o->fini_listener;
	    $o->scanner_stop($_) for $o->slots_idle;
	    $o->refresh_fds;
	    goto PING_RESPONSE;
	}
    }
    elsif ($cmd eq 'R') {
	$o->log("INFO", "received *soft* reset request. Resetting!");
	$o->fini_scanners;
	$o->init_scanners;
	$o->init;
	$o->chat_insert($ent); # re-insert this packet into the chat queue.
	goto PING_RESPONSE;
    }
    elsif ($cmd eq 'r') {
	if ($kv{hard}) { # hard respawn -- kill any running clients
	    $o->log("INFO", "received *hard* respawn request. Restarting scanners!");
	    for my $slot ($o->slots_inuse) {
		my $ent = $o->slot($slot);
		$ent->{resp} = [
		    problem => 'RESPAWN',
		    descrip => 'Scanner respawned explicitly',
		    command => $ent->{cmd},
		    file    => $ent->{file},
		    created => $ent->{create},
		    started => $ent->{start},
		    elapsed => time() - $ent->{start},
		    report  => $ent->{rbuf},
		    op	    => 'e',
		];
		$ent->{read_scanner} = 0;
		$ent->{write_scanner} = 0;
		$ent->{write_sock} = 1;
		$ent->{read_sock} = 1;
		$ent->{restart} = 1;
		$o->free_slot($slot);
	    }
	    $o->fini_scanners;
	    $o->init_scanners;
	}
	else {
	    $o->log("INFO", "received *soft* respawn request. Scheduling respawns!");
	    $o->slot($_)->{restart} = 1 for $o->slots_inuse;
	    $o->scanner_restart($_) for $o->slots_idle;
	    for ($o->slots_dead) {
		$o->slot_dead($_, undef); # revive this slot
		$o->scanner_start($_);    # start it up (may kill it again)
	    }
	    $o->refresh_fds;
	}
	goto PING_RESPONSE;
    }
}

sub read_sock {
    my ($o, $ent) = @_;

    # Count the number of currently-running commands under this socket.
    my $count = grep { $_ && $_->{sock} == $ent->{sock} } $o->chatting;

    # Read several many packets from the socket
    while (1) {
	my ($op, @kv) = recv($ent);

	# Check for a valid packet
	if (defined $op) {
	    $o->log("DEBUG", "Request '$op' from client ${$ent->{sock}}");
	    my $new = $o->alloc_ent($ent->{sock});
	    $new->{eof} = 1;
	    $o->chat_insert($new);
	    $o->handle($op, $new, @kv);
	    $count++;
	    next;
	}
	elsif ($count <= 1) {
	    $o->log("DEBUG", "Client ${$ent->{sock}} finished");
	    $o->free_queued($ent);
	    $o->chat_free($ent);
	}
	last;
    }
}

sub write_sock {
    my ($o, $ent) = @_;
    my @kv = @{$ent->{resp}} or return;
    $o->log("DEBUG", "Writing scan results to ${$ent->{sock}}");
    my $op = {@kv}->{op};
    eval { send($ent, $op, @kv, backends => scalar @{$o->{backends}}) };
    $o->chat_free($ent) if $@;
    $ent->{del} = 1;
}

sub refresh_fds {
    my $o = shift;
    $o->{RFDS} = [$o->{pool}->rfds];
    $o->{WFDS} = [$o->{pool}->wfds];
    $o->{EFDS} = [$o->{pool}->efds];
}

#=============================================================================
# Queue management. queue() adds a new request to the queued requests;
# dequeue() removes the first request from the Queue and inserts it into the
# specified slot; free_queued() searches through the queue for entities whose
# {sock} entry is the same as the given request, and deletes them.
#=============================================================================
sub enqueue {
    my ($o, $ent) = @_;
    $ent->{start} = time;
    push @{$o->{pending}}, $ent;
}

sub dequeue {
    my ($o, $slot) = @_;
    my $ent;
    $ent = shift @{$o->{pending}} while $o->queuesize and not defined $ent;
    return unless $ent;
    #$ent->{start} = time;
    $o->{running}[$slot] = $ent;
}

sub popqueue {
    my $o = shift;
    my $ent;
    $ent = shift @{$o->{pending}} while $o->queuesize and not defined $ent;
    $ent;
}

sub queuesize { my $o = shift; return scalar @{$o->{pending}} }

sub queued { my $o = shift; @{$o->{pending}} }

sub free_queued {
    my ($o, $ent) = @_;
    my $fileno = $ent->{sock}->fileno;
    $o->{pending} = [(
	grep { $_ && $_->{sock}->fileno != $fileno }
	@{$o->{pending}}
    )];
}

#=============================================================================
# Mux management. We have a fixed-size array of scanners connected to us via
# pipes. These subs check for an empty slot, free a slot, etc.
#=============================================================================
sub slot {
    my ($o, $slot) = @_;
    return $o->{running}[$slot];
}

sub slot_inuse {
    my ($o, $slot) = @_;
    return defined $o->{running}[$slot];
}

sub slots_inuse {
    my $o = shift;
    grep { $o->slot_inuse($_) } $o->slots;
}

sub slot_dead {
    my ($o, $slot, $dead) = @_;
    if (@_ == 3 and $dead) {
	my $respawn_time = time() + $my_config{daemon}{respawn_wait};
	$o->{dead_slots}[$slot] = $respawn_time;
    }
    elsif (@_ == 3) {
	$o->{dead_slots}[$slot] = undef;
    }
    $o->{dead_slots}[$slot];
}

sub slots_dead {
    my $o = shift;
    grep { $o->slot_dead($_) } $o->slots;
}

sub check_dead_slots {
    my ($o, $now) = @_;
    for my $slot ($o->slots) {
	my $revive_time = $o->slot_dead($slot) or next;
        if ($now >= $revive_time) {
	    $o->log("DEBUG", "restarting dead slot $slot");
            $o->scanner_start($slot);
            $o->refresh_fds;
            $o->slot_dead($slot, undef);
        }
        else {
	    # don't log anything here -- too noisy
        }
    }
}

sub slots_idle {
    my $o = shift;
    grep { !$o->slot_inuse($_) and !$o->slot_dead($_) } $o->slots;
}

sub find_free {
    my $o = shift;
    for my $slot ($o->slots) {
        next if $o->slot_inuse($slot);
        next if $o->slot_dead($slot);
        return $slot;
    }
    undef;
}

sub free_slot {
    my ($o, $slot, $warn) = @_;
    my $restart = $o->{running}[$slot]{restart};
    $o->{running}[$slot] = undef;

    # Increment the use count of this slot
    my $reuse = $my_config{daemon}{reinit_limit};
    if ($restart or $reuse and $o->{pool}->used($slot) > $reuse) {
	$o->log("DEBUG", "restarting dead slot $slot");
        $o->scanner_restart($slot, $warn);
        $o->refresh_fds();
    }
}

#=============================================================================
# Create/destroy "entities". These are basically request objects; you get one
# for each packet sent to the daemon.
#=============================================================================
sub alloc_ent {
    my ($o, $sock) = @_;
    return {
        sock => $sock,          # client socket
        cmd => '',              # actual command
        resp => [],             # response to the client
        rbuf => '',             # has been read from scanner
        read_sock  => 1,        # okay to read from socket
        read_scanner => 0,      # okay to read from scanner
        write_sock => 0,        # okay to write to socket
        write_scanner => 0,     # okay to write to scanner
        create => time(),       # to detect timeouts
        eof => 0,               # EOF on the socket
	restart => 0,
    };
}

#=============================================================================
# Handle select()-related stuff. File numbers. Vectors. Bits.
#=============================================================================
sub fno {
    my $fd = shift;
    return $fd unless defined $fd;
    return $fd->fileno if eval { no warnings; $fd->can("fileno"); 1 };
    return $fd;
}

sub fd_set {
    my $vec = '';
    for my $fd (@_) {
        next unless defined $fd and fno($fd);
        vec($vec, fno($fd), 1) = 1;
    }
    return $vec;
}

sub fd_isset {
    my $vec = shift;
    my $fd  = shift;
    return 0 unless defined $fd;
    return vec($vec, fno($fd), 1);
}

#=============================================================================
# Errors.
#=============================================================================
sub err_scanner_closed {
    my ($o, $slot, $ent) = @_;
    my $elapsed = time - $ent->{start};
    my $name    = $o->slot_name($slot);
    my $descrip = "Scanner $slot ($name) closed unexpectedly";
    $o->log("ERROR", $descrip);
    $ent->{resp} = [
	op	=> 'e',
        problem => 'EOF',
        descrip => $descrip,
        command => $ent->{cmd},
        file    => $ent->{file},
        created => $ent->{create},
        started => $ent->{start},
        elapsed => $elapsed,
        report  => $ent->{rbuf},
    ];
    $ent->{read_scanner} = 0;
    $ent->{write_scanner} = 0;
    $ent->{write_sock} = 1;
    $ent->{read_sock} = 1;
    $ent->{restart} = 1;
    $o->free_slot($slot, 1);
    return;
}

#=============================================================================
# Wrap those ProcessPool methods which can throw exceptions on errors.
# Respond by marking that slot as dead.
#=============================================================================
sub slot_name {
    my ($o, $slot) = @_;
    return $o->{backends}[$slot % @{$o->{backends}}];
}

sub scanner_start {
    my ($o, $slot) = @_;
    my $backend = $o->slot_name($slot);
    my $impl = $o->{bk}{$backend};
    $o->{pool}->start($slot, sub {
	my %kv = @_ or exit(1);
	my ($scan, $clean) = @$impl;
	if ($kv{action} eq 'scan') {
	    delete $kv{action};				#change
	    my @v = &$scan($backend, %kv);		#change
	    @v = map { (v => $_) } @v;
	    return (file => $kv{file}, @v);
	}
	elsif ($kv{action} eq 'clean') {
	    my $status;
	    if ($clean) {
		$status = &$clean($backend, $kv{file});
	    }
	    else {
	        delete $kv{action};
   	        my @v = &$scan($backend, %kv);		#change
		$status = @v ? 'infected' : 'novirus';	#change
	    }
	    return (file => $kv{file}, status => $status);
	}
    });
}

sub scanner_restart {
    my ($o, $slot, $warn) = @_;
    $o->scanner_stop($slot, $warn);
    $o->scanner_start($slot);
}

sub scanner_stop {
    my ($o, $slot, $warn) = @_;
    my ($status, $signal) = $o->{pool}->stop($slot);
    if ($signal or $status) {
	my $name = $o->slot_name($slot);
	$o->log("INFO", "Scanner $slot ($name) exited uncleanly: " .
			"exit=$status signal=$signal")
	    if $warn;
    }
}

sub rfd {
    my $o = shift;
    my $i = shift;
    $o->{RFDS}[$i];
}

sub wfd {
    my $o = shift;
    my $i = shift;
    $o->{WFDS}[$i];
}

sub efd {
    my $o = shift;
    my $i = shift;
    $o->{EFDS}[$i];
}

# Returns 1 when there is at least one daemon-capable backend enabled but none
# running in daemon mode.
sub no_daemons {
    my $backends = 0;

    for (grep $my_config{backend}{$_}{enabled}, keys %{$my_config{backend}}) {
	$backends++ if defined $my_config{backend}{$_}{daemon};
	if ($my_config{backend}{$_}{daemon}) {
	    # We found an enabled backend running in daemon mode, so return.
	    return 0;
	}
    }
    return $backends > 0;
}

1;
